/*
  Copyright (c) 2020 gotomaxx GmbH. All rights reserved.
  See the LICENSE.txt in the root folder of the source tree.
*/

#include <Shlwapi.h>
#include <atlbase.h>
#include <iostream>
#include <string>

static const int ErrorArguments = 1;
static const int ErrorUnexpected = 2;
static const int ErrorCreateObject = 3;
static const int ErrorLoadXSLT = 4;
static const int ErrorLoadXML = 5;
static const int ErrorTransformXML = 6;
static const int ErrorNodeSelection = 7;

int validate(const std::wstring& xmlFile, const std::wstring& validationFile)
{
  HRESULT hr;
  CComQIPtr<IXMLDOMDocument> validationXSLT;
  hr = CoCreateInstance(CLSID_DOMDocument, nullptr, CLSCTX_INPROC_SERVER,
                        IID_PPV_ARGS(&validationXSLT));
  if (FAILED(hr))
  {
    std::wcout << "Error: Create XML document object failed\n";
    return ErrorCreateObject;
  }

  validationXSLT->put_async(VARIANT_FALSE);
  VARIANT_BOOL isSuccessful = VARIANT_FALSE;
  hr = validationXSLT->load(CComVariant(validationFile.c_str()), &isSuccessful);
  if (FAILED(hr) || !isSuccessful)
  {
    std::wcout << "Error: Loading validation XSLT failed\n";
    return ErrorLoadXSLT;
  }

  CComQIPtr<IXMLDOMDocument> dataXML;
  hr = CoCreateInstance(CLSID_DOMDocument, nullptr, CLSCTX_INPROC_SERVER, IID_PPV_ARGS(&dataXML));
  dataXML->put_async(VARIANT_FALSE);
  hr = dataXML->load(CComVariant(xmlFile.c_str()), &isSuccessful);
  if (FAILED(hr) || !isSuccessful)
  {
    std::wcout << "Error: Loading invoice XML failed\n";
    return ErrorLoadXML;
  }

  CComQIPtr<IXMLDOMDocument> validationResultXML;
  hr = CoCreateInstance(CLSID_DOMDocument, nullptr, CLSCTX_INPROC_SERVER,
                        IID_PPV_ARGS(&validationResultXML));

  if (FAILED(hr))
  {
    std::wcout << "Error: Create XML document object failed\n";
    return ErrorCreateObject;
  }
  validationResultXML->put_async(VARIANT_FALSE);

  hr = dataXML->transformNodeToObject(validationXSLT, CComVariant(validationResultXML));
  if(FAILED(hr))
  {
    std::wcout << "Error: Transformation failed\n";
    return ErrorTransformXML;
  }

  CComQIPtr<IXMLDOMNodeList> errorNodes;
  hr = validationResultXML->selectNodes(CComBSTR(L"/svrl:schematron-output/svrl:failed-assert"),
                                        &errorNodes);
  if (FAILED(hr))
  {
    std::wcout << "Error: Node selection failed\n";
    return ErrorNodeSelection;
  }

  long errorCount = -1;
  hr = errorNodes->get_length(&errorCount);
  std::wcout << "Results: " << errorCount << "\n";
  for (long errorIndex = 0; errorIndex < errorCount; errorIndex++)
  {
    CComQIPtr<IXMLDOMNode> errorNode;
    hr = errorNodes->get_item(errorIndex, &errorNode);
    if (!errorNode) continue;

    CComQIPtr<IXMLDOMNamedNodeMap> attributes;
    hr = errorNode->get_attributes(&attributes);
    if (!attributes) continue;

    CComBSTR flagStr, textStr;

    CComQIPtr<IXMLDOMNode> flagNode;
    attributes->getNamedItem(CComBSTR(L"flag"), &flagNode);
    if (flagNode) flagNode->get_text(&flagStr);

    CComQIPtr<IXMLDOMNode> textNode;
    CComBSTR nodeName;
    hr = errorNode->get_firstChild(&textNode);
    if (textNode && textNode->get_nodeName(&nodeName) == S_OK &&
        wcscmp(nodeName, L"svrl:text") == 0)
    { textNode->get_text(&textStr); }

    std::wcout << " " << (flagStr ? (const wchar_t*)flagStr : L"<unknown>") << ": "
               << (textStr ? (const wchar_t*)textStr : L"<unknown>") << "\n";
  }

  return 0;
}

int wmain(int ac, wchar_t* av[], wchar_t* envp[])
{
  std::wcout << "gotomaxx eInvoice Validator | Copyright (c) 2020 gotomaxx GmbH, All Rights Reserved\n";
  if (ac != 3)
  {
    std::wcout << "Error: invalid arguments. Expected cii|ubl|facturx_extended xml-file\n";
    return ErrorArguments;
  }

  std::wstring profile = av[1];
  if (profile != L"cii" && profile != L"ubl" && profile != L"facturx_extended")
  {
    std::wcout << "Error: invalid profile. Expected cii|ubl|facturx_extended\n";
    return ErrorArguments;
  }

  std::wstring xmlFile = av[2];
  if (!PathFileExistsW(xmlFile.c_str()))
  {
    std::wcout << "Error: input file does not exist\n";
    return ErrorArguments;
  }

  wchar_t* filePathEnd = PathFindFileNameW(av[0]);
  std::wstring validationFile(av[0], filePathEnd);
  if (profile == L"cii")
    validationFile += L"\\EN16931-CII-Validation.xslt";
  else if (profile == L"ubl")
    validationFile += L"\\EN16931-UBL-Validation.xslt";
  else if (profile == L"facturx_extended")
    validationFile += L"\\FACTUR-X_EXTENDED-Validation.xslt";

  HRESULT hr;
  hr = CoInitialize(nullptr);
  if (FAILED(hr))
  {
    std::wcout << "Error: COM initialization failed\n";
    return ErrorUnexpected;
  }

  int ret = validate(xmlFile, validationFile);

  CoUninitialize();

  return ret;
}
